#include <kernel/driver/pci.h>
#include <kernel/driver/nic/ne2k.h>
#include <kernel/ibm_i386.h>
#include <errno.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------------------
int
ne2k_reset (uintptr_t io, void *address)
{
  int     status;
  int     i;
  uint8_t sa_prom[NE2K_SAPROM_SIZE];
  uint8_t par[6];
  //
  // -----------------------------------------------------------------
  //  RESET
  // -----------------------------------------------------------------
  //
  status = in_8 (io + NE2K_RESET);
  out_8 ((io + NE2K_RESET), 0xFF);
  out_8 ((io + NE2K_RESET), status);
  //
  // Interrupt status register (ISR)
  // .-------------------------------.
  // |RST|RDC|CNT|OVW|TXE|RXE|PTX|PRX|
  // |-------------------------------|
  // | 1 | ? | ? | ? | ? | ? | ? | ? | 0x80
  // `-------------------------------'
  //   |
  //   Reset status
  //
  // Verify to have reset the NIC.
  //
  status = ne2k_isr_expect (io, 0x80);
  if (status)
    {
      errset (errno);
      return (-1);
    }
  //
  // Reset all ISR register flags.
  //
  out_8 ((io + NE2K_ISR), 0xFF);
  //
  // -----------------------------------------------------------------
  //  GET ETHERNET ADDRESS FROM SA-PROM (Station Address PROM)
  // -----------------------------------------------------------------
  //
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 1 | 0 | 0 | 0 | 0 | 1 | 0x21
  // `-------------------------------'
  //  \_____/ \_________/          |
  //     |         |               STOP
  //     |         |
  //     |         Abort/complete
  //     |         remote DMA
  //     |
  //     Register
  //     page 0
  //
  out_8 ((io + NE2K_CR), 0x21);
  //
  // Interrupt status register (ISR)
  // .-------------------------------.
  // |RST|RDC|CNT|OVW|TXE|RXE|PTX|PRX|
  // |-------------------------------|
  // | 1 | ? | ? | ? | ? | ? | ? | ? | 0x80
  // `-------------------------------'
  //   |
  //   Reset status
  //
  // Verify to have reset the NIC.
  //
  status = ne2k_isr_expect (io, 0x80);
  if (status)
    {
      errset (errno);
      return (-1);
    }
  //
  // Data configuration register (DCR)
  // .-------------------------------.
  // | - |FT1|FT0|ARM| LS|LAS|BOS|WTS|
  // |-------------------------------|
  // | 0 | 1 | 0 | 0 | 1 | 0 | 0 | 0 | 0x48
  // `-------------------------------'
  //      \_____/  :   |   :   :   :
  //         |     :   |   :   :  Byte DMA transfer
  //         |     :   |   :   :
  //         |     :   |   :   Little endian byte order
  //         |     :   |   :
  //         |     :   |  Dual 16 bit DMA mode
  //         |     :   |
  //         |     :   Loopback OFF (normal operation)
  //         |     :
  //         |    Send Command non executed: all frames removed from
  //         |    Buffer Ring under program control
  //         |
  //         FIFO threshold 8 bytes
  //
  out_8 ((io + NE2K_DCR), 0x48);
  //
  // Reset remote byte count registers.
  //
  out_8 ((io + NE2K_RBCR0), 0x00);
  out_8 ((io + NE2K_RBCR1), 0x00);
  //
  // Disable interrupts with an empty mask.
  //
  out_8 ((io + NE2K_IMR), 0x00);
  //
  // Reset all ISR register flags.
  //
  out_8 ((io + NE2K_ISR), 0xFF);
  //
  // Receive configuration register (RCR)
  // .-------------------------------.
  // | - | - |MON|PRO| AM| AB| AR|SEP|
  // |-------------------------------|
  // | 0 | 0 | 1 | 0 | 0 | 0 | 0 | 0 | 0x20
  // `-------------------------------'
  //           |   :   :   :   :   :
  //           |   :   :   :   :   Frames with receive errors are
  //           |   :   :   :   :   rejected
  //           |   :   :   :   :
  //           |   :   :   :   Frames with fewer than 64 bytes rejected
  //           |   :   :   :
  //           |   :   :   Frames with broadcast destination rejected
  //           |   :   :   accepted
  //           |   :   :
  //           |   :   Frames with multicast destination address
  //           |   :   not checked
  //           |   :
  //           |   Physical address of node must match the station
  //           |   address
  //           |
  //           Monitor mode: frames checked but not buffered to memory
  //
  //  Monitor mode.
  //
  out_8 ((io + NE2K_RCR), 0x20);
  //
  // Transmit configuration register (TCR)
  // .--------------------------------.
  // | - | - | - |OFST|ATD|LB1|LB0|CRC|
  // |--------------------------------|
  // | 0 | 0 | 0 | 0  | 0 | 0 | 1 | 0 | 0x02
  // `--------------------------------'
  //                       \_____/  :
  //                          |     CRC appended by the transmitter
  //                          |
  //                          Internal loopback (mode 1)
  //
  // [Loopback is not supported by Bochs]
  //
  // Transmit loopback.
  //
  out_8 ((io + NE2K_TCR), 0x02);
  //
  // Remote byte count registers to NE2K_SAPROM_SIZE: the bytes to be
  // read from SA-PROM.
  //
  out_8 ((io + NE2K_RBCR0), NE2K_SAPROM_SIZE);
  out_8 ((io + NE2K_RBCR1), (NE2K_SAPROM_SIZE >> 8));
  //
  // Set the remote DMA address to zero.
  //
  out_8 ((io + NE2K_RSAR0), 0x00); // Must be zero.
  out_8 ((io + NE2K_RSAR1), 0x00);
  //
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 0 | 0 | 1 | 0 | 1 | 0 | 0x0A
  // `-------------------------------'
  //   \___/   \_______/       |
  //     |         |           START
  //     |         Read
  //     |
  //     Register
  //     page 0
  //
  out_8 ((io + NE2K_CR), 0x0A);
  //
  // Save the SA-PROM content.
  //
  for (i = 0; i < NE2K_SAPROM_SIZE; i++)
    {
      sa_prom[i] = in_8 (io + NE2K_DATA);
    }
  //
  // Set NIC physical address from SA-PROM data.
  //
  par[0] = sa_prom[0];
  par[1] = sa_prom[2];
  par[2] = sa_prom[4];
  par[3] = sa_prom[6];
  par[4] = sa_prom[8];
  par[5] = sa_prom[10];
  //
  // Copy to the `address' pointer.
  //
  if (address != NULL)
    {
      memcpy (address, par, (size_t) 6);
    }
  //
  // -----------------------------------------------------------------
  //  INITIALIZATION SEQUENCE
  // -----------------------------------------------------------------
  //
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 1 | 0 | 0 | 0 | 0 | 1 | 0x21
  // `-------------------------------'
  //  \_____/ \_________/          |
  //     |         |               STOP
  //     |         |
  //     |         Abort/complete
  //     |         remote DMA
  //     |
  //     Register
  //     page 0
  //
  out_8 ((io + NE2K_CR), 0x21);
  //
  // There is no need to check the ISR value. At this point,
  // ISR might report a reset status or a remote DMA complete.
  // Go to the DCR register.
  //
  // Data configuration register (DCR)
  // .-------------------------------.
  // | - |FT1|FT0|ARM| LS|LAS|BOS|WTS|
  // |-------------------------------|
  // | 0 | 1 | 0 | 0 | 1 | 0 | 0 | 0 | 0x48
  // `-------------------------------'
  //      \_____/  :   |   :   :   :
  //         |     :   |   :   :  Byte DMA transfer
  //         |     :   |   :   :
  //         |     :   |   :   Little endian byte order
  //         |     :   |   :
  //         |     :   |  Dual 16 bit DMA mode
  //         |     :   |
  //         |     :   Loopback OFF (normal operation)
  //         |     :
  //         |    Send Command non executed: all frames removed from
  //         |    Buffer Ring under program control
  //         |
  //         FIFO threshold 8 bytes
  //
  out_8 ((io + NE2K_DCR), 0x48);
  //
  // Reset remote byte count registers.
  //
  out_8 ((io + NE2K_RBCR0), 0x00);
  out_8 ((io + NE2K_RBCR1), 0x00);
  //
  // Receive configuration register (RCR)
  // .-------------------------------.
  // | - | - |MON|PRO| AM| AB| AR|SEP|
  // |-------------------------------|
  // | 0 | 0 | 0 | 0 | 0 | 1 | 0 | 0 | 0x04
  // `-------------------------------'
  //           :   :   :   |   :   :
  //           :   :   :   |   :   Frames with receive errors are
  //           :   :   :   |   :   rejected
  //           :   :   :   |   :
  //           :   :   :   |   Frames with fewer than 64 bytes rejected
  //           :   :   :   |
  //           :   :   :   Frames with broadcast destination address
  //           :   :   :   accepted
  //           :   :   :
  //           :   :   Frames with multicast destination address
  //           :   :   not checked
  //           :   :
  //           :   Physical address of node must match the station
  //           :   address
  //           :
  //           Frames buffered to memory
  //
  //  Normal operation and broadcast accepted.
  //
  out_8 ((io + NE2K_RCR), 0x04);
  //
  // Transmit page start (local DMA).
  //
  out_8 ((io + NE2K_TPSR), NE2K_TX_START);
  //
  // Transmit configuration register (TCR)
  // .--------------------------------.
  // | - | - | - |OFST|ATD|LB1|LB0|CRC|
  // |--------------------------------|
  // | 0 | 0 | 0 | 0  | 0 | 0 | 1 | 0 | 0x02
  // `--------------------------------'
  //                       \_____/  :
  //                          |    CRC appended by the transmitter
  //                          |
  //                          Internal loopback (mode 1)
  //
  // [Loopback is not supported by Bochs]
  //
  // Transmit loopback.
  //
  out_8 ((io + NE2K_TCR), 0x02);
  //
  // Set receive buffer page start (local DMA).
  //
  out_8 ((io + NE2K_PSTART), NE2K_RX_START);
  //
  // Set boundary: the frame not yet read. At the moment, it is the same
  // as the receive buffer page start.
  //
  out_8 ((io + NE2K_BNRY), NE2K_RX_START);
  //
  // Set receive buffer page stop (local DMA).
  //
  out_8 ((io + NE2K_PSTOP), NE2K_RX_STOP);
  //
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 1 | 1 | 0 | 0 | 0 | 0 | 1 | 0x61
  // `-------------------------------'
  //  \_____/ \_________/          |
  //     |         |               STOP
  //     |         |
  //     |         Abort/complete
  //     |         remote DMA
  //     |
  //     Register
  //     page 1
  //
  out_8 ((io + NE2K_CR), 0x61);
  //
  // Save physical address and multicast address.
  //
  out_8 ((io + NE2K_PAR0), par[0]);
  out_8 ((io + NE2K_PAR1), par[1]);
  out_8 ((io + NE2K_PAR2), par[2]);
  out_8 ((io + NE2K_PAR3), par[3]);
  out_8 ((io + NE2K_PAR4), par[4]);
  out_8 ((io + NE2K_PAR5), par[5]);
  //
  out_8 ((io + NE2K_MAR0), 0);
  out_8 ((io + NE2K_MAR1), 0);
  out_8 ((io + NE2K_MAR2), 0);
  out_8 ((io + NE2K_MAR3), 0);
  out_8 ((io + NE2K_MAR4), 0);
  out_8 ((io + NE2K_MAR5), 0);
  out_8 ((io + NE2K_MAR6), 0);
  out_8 ((io + NE2K_MAR7), 0);
  //
  // Set current page: the first frame to be saved inside the receive
  // buffer. At the moment, it is the same as the buffer page start.
  //
  out_8 ((io + NE2K_CURR), NE2K_RX_START);
  //
  // Command register (CR)
  // .-------------------------------.
  // |PS1|PS0|RD2|RD1|RD0|TXP|STA|STP|
  // |-------------------------------|
  // | 0 | 0 | 1 | 0 | 0 | 0 | 1 | 0 | 0x22
  // `-------------------------------'
  //  \_____/ \_________/      |
  //     |         |           START
  //     |         |
  //     |         Abort/complete
  //     |         remote DMA
  //     |
  //     Register
  //     page 0
  //
  out_8 ((io + NE2K_CR), 0x22);
  //
  // Reset all ISR register flags.
  //
  out_8 ((io + NE2K_ISR), 0xFF);
  //
  // ISR will be polled, but received packets will fire the IRQ,
  // although it is not necessary. So the IMR (Interrupt mask register)
  // is now set properly with the value 0x01.
  //
  // Interrupt mask register (IMR)
  // .---------------------------------------.
  // | -- |RDCE|CNTE|OVWE|TXEE|RXEE|PTXE|PRXE|
  // |---------------------------------------|
  // |  0 |  0 |  0 |  0 |  0 |  0 |  0 |  1 | 0x01
  // `---------------------------------------'
  //                                       |
  //                                       Enable interrupt when packet
  //                                       received
  //
  out_8 ((io + NE2K_IMR), 0x01);
  //
  // Transmit configuration register (TCR)
  // .--------------------------------.
  // | - | - | - |OFST|ATD|LB1|LB0|CRC|
  // |--------------------------------|
  // | 0 | 0 | 0 | 0  | 0 | 0 | 0 | 0 | 0x00
  // `--------------------------------'
  //
  // Normal operation.
  //
  out_8 ((io + NE2K_TCR), 0x00);
  //
  return (0);
}
