#ifndef _KERNEL_IBM_I386_H
#define _KERNEL_IBM_I386_H   1
//----------------------------------------------------------------------
#include <stdint.h>
#include <inttypes.h>
#include <stdbool.h>
#include <stdarg.h>
//----------------------------------------------------------------------
// GDT
//----------------------------------------------------------------------
#define GDT_ITEMS       256    // Max is 8192 items.
//
typedef struct {
    uint32_t  limit_a              : 16,
              base_a               : 16;
    uint32_t  base_b               : 8,
              accessed             : 1,
              write_execute        : 1,
              expansion_conforming : 1,
              code_or_data         : 1,
              code_data_or_system  : 1,
              dpl                  : 2,
              present              : 1,
              limit_b              : 4,
              available            : 1,
              reserved             : 1,
              big                  : 1,
              granularity          : 1,
              base_c               : 8;
} gdt_t;
//
extern gdt_t gdt_table[GDT_ITEMS];
//----------------------------------------------------------------------
typedef struct {
    uint16_t  limit;
    uint32_t  base;
} __attribute__ ((packed)) gdtr_t;      // [1]
//
extern gdtr_t gdt_register;
//
// [1] It is necessary that the structure be compact, so that it
//     uses exactly 48 bits. That is why the attribute `packed'
//     for the GNU C compiler is used.
//----------------------------------------------------------------------
int gdt_segment (int segment, uint32_t base, uint32_t limit,
                 bool present, bool code, unsigned char dpl);
//
void gdt_print    (void *gdtr, unsigned int first, unsigned int last);
void gdt_load     (void *gdtr);
void gdt          (void);
//
// Segment 0 is not used,
// segment 1 is for kernel code,
// segment 2 is for kernel data,
// segment 3 is for process 1 code,
// segment 4 is for process 1 data,
// ...
//
#define gdt_pid_to_segment_text(p)  (p*2+1)
#define gdt_pid_to_segment_data(p)  (p*2+2)
#define gdt_segment_text_to_pid(s)  (s/2)
#define gdt_segment_data_to_pid(s)  (s/2-1)
//----------------------------------------------------------------------
// IDT
//----------------------------------------------------------------------
#define IDT_ITEMS      129      // 0-128 0x00-0x80
//----------------------------------------------------------------------
typedef struct {
    uint32_t  offset_a : 16,
              selector : 16;
    uint32_t  filler   : 8,
              type     : 4,
              system   : 1,
              dpl      : 2,
              present  : 1,
              offset_b : 16;
} idt_t;
//
extern idt_t idt_table[IDT_ITEMS];
//----------------------------------------------------------------------
typedef struct {
    uint16_t  limit;
    uint32_t  base;
} __attribute__ ((packed)) idtr_t;
//
extern idtr_t idt_register;
//----------------------------------------------------------------------
void idt_descriptor (int desc, void *isr, uint16_t selector,
                     bool present, char type, char dpl);
void idt_load       (void *idtr);
void idt            (void);
void idt_irq_remap  (unsigned int offset_1, unsigned int offset_2);
void idt_print      (void *idtr, unsigned int first, unsigned int last);
//----------------------------------------------------------------------
// ISR
//----------------------------------------------------------------------
void  isr_0   (void);
void  isr_1   (void);
void  isr_2   (void);
void  isr_3   (void);
void  isr_4   (void);
void  isr_5   (void);
void  isr_6   (void);
void  isr_7   (void);
void  isr_8   (void);
void  isr_9   (void);
void  isr_10  (void);
void  isr_11  (void);
void  isr_12  (void);
void  isr_13  (void);
void  isr_14  (void);
void  isr_15  (void);
void  isr_16  (void);
void  isr_17  (void);
void  isr_18  (void);
void  isr_19  (void);
void  isr_20  (void);
void  isr_21  (void);
void  isr_22  (void);
void  isr_23  (void);
void  isr_24  (void);
void  isr_25  (void);
void  isr_26  (void);
void  isr_27  (void);
void  isr_28  (void);
void  isr_29  (void);
void  isr_30  (void);
void  isr_31  (void);
void  isr_32  (void);
void  isr_33  (void);
void  isr_34  (void);
void  isr_35  (void);
void  isr_36  (void);
void  isr_37  (void);
void  isr_38  (void);
void  isr_39  (void);
void  isr_40  (void);
void  isr_41  (void);
void  isr_42  (void);
void  isr_43  (void);
void  isr_44  (void);
void  isr_45  (void);
void  isr_46  (void);
void  isr_47  (void);
void  isr_128 (void);
//
char *isr_exception_name (int exception);
//
void  isr_exception_unrecoverable (uint32_t eax, uint32_t ecx,
                                   uint32_t edx, uint32_t ebx,
                                   uint32_t ebp, uint32_t esi,
                                   uint32_t edi, uint32_t ds,
                                   uint32_t es,  uint32_t fs,
                                   uint32_t gs,  uint32_t interrupt,
                                   uint32_t error, uint32_t eip,
                                   uint32_t cs, uint32_t eflags);
//
void  isr_irq_clear (uint32_t idtn);
void  isr_irq_clear_pic1 (void);
void  isr_irq_clear_pic2 (void);
//----------------------------------------------------------------------
// I/O
//----------------------------------------------------------------------
uint32_t _in_8    (uint32_t port);
uint32_t _in_16   (uint32_t port);
uint32_t _in_32   (uint32_t port);
void     _out_8   (uint32_t port, uint32_t value);
void     _out_16  (uint32_t port, uint32_t value);
void     _out_32  (uint32_t port, uint32_t value);
//
#define in_8(port)           ((unsigned int) _in_8 ((uint32_t) port))
#define in_16(port)          ((unsigned int) _in_16 ((uint32_t) port))
#define in_32(port)          ((unsigned int) _in_32 ((uint32_t) port))
#define out_8(port, value)   (_out_8 ((uint32_t) port, \
                                      (uint32_t) value))
#define out_16(port, value)  (_out_16 ((uint32_t) port, \
                                       (uint32_t) value))
#define out_32(port, value)  (_out_32 ((uint32_t) port, \
                                       (uint32_t) value))
//----------------------------------------------------------------------
// Interrupt on/off
//----------------------------------------------------------------------
void     cli     (void);
void     sti     (void);
void     irq_on  (unsigned int irq);
void     irq_off (unsigned int irq);
//----------------------------------------------------------------------
#endif
