#include <errno.h>
#include <kernel/proc.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------------------
void
s__exit (pid_t pid, int status)
{
  pid_t           child;
  pid_t           parent = proc_table[pid].ppid;
  int             proc_count;
  pid_t           extra;
  int             sigchld = 0;
  int             fdn;
  tty_t          *tty;
  int             closed;
  fd_t           *fd;
  //
  proc_table[pid].status = PROC_ZOMBIE;
  proc_table[pid].ret = status;
  proc_table[pid].sig_status = 0;
  proc_table[pid].sig_ignore = 0;
  //
  // Close files.
  //
  for (fdn = 0; fdn < OPEN_MAX; fdn++)
    {
      closed = s_close (pid, fdn);
      //
      // Close might fail for work in progress.
      //
      if (closed < 0 && errno == EINPROGRESS)
        {
          //
          // Should be a socket, but close badly,
          // because we cannot wait.
          //
          fd = fd_reference (pid, &fdn);
          if (fd->file->sock != NULL)
            {
              fd->file->sock->active = 0;
              s_close (pid, fdn);
            }
        }
    }
  //
  // Close current directory.
  //
  inode_put (proc_table[pid].inode_cwd);
  //
  // Close the controlling terminal, if it is a process leader with
  // such a terminal.
  //
  if (proc_table[pid].pgrp == pid && proc_table[pid].device_tty != 0)
    {
      tty = tty_reference (proc_table[pid].device_tty);
      //
      // Verify.
      //
      if (tty == NULL)
	{
	  //
	  // Show a kernel message.
	  //
	  k_printf ("kernel alert: cannot find the terminal item "
		    "for device 0x%04x!\n",
		    (int) proc_table[pid].device_tty);
	}
      else if (tty->pgrp != pid)
	{
	  //
	  // Show a kernel message.
	  //
	  k_printf ("kernel alert: terminal device 0x%04x should "
		    "be associated to the process group %i, but it "
		    "is instead related to process group %i!\n",
		    (int) proc_table[pid].device_tty, (int) pid,
		    (int) tty->pgrp);
	}
      else
	{
	  tty->pgrp = 0;
	}
    }
  //
  // Data and text might share the same address space. If they are
  // are on different places, then must verify if the text is not
  // shared by other processes.
  //
  if (proc_table[pid].domain_data == 0)
    {
      //
      // Text and data are together.
      //
      mb_free (proc_table[pid].address_text,
	       (proc_table[pid].domain_text
		+ proc_table[pid].extra_data));
    }
  else
    {
      //
      // Data is separate and is to be removed alone.
      //
      mb_free (proc_table[pid].address_data,
	       (proc_table[pid].domain_data
		+ proc_table[pid].extra_data));
      //
      // Now must verify if no other process uses the same text
      // memory.
      //
      for (proc_count = 0, extra = 0; extra < PROCESS_MAX; extra++)
	{
	  if (proc_table[extra].status == PROC_EMPTY ||
	      proc_table[extra].status == PROC_ZOMBIE)
	    {
	      continue;
	    }
	  if (proc_table[pid].address_text
	      == proc_table[extra].address_text)
	    {
	      proc_count++;
	    }
	}
      if (proc_count == 0)
	{
	  //
	  // The code segment can be released, because no other
	  // process, except the current one (to be closed),
	  // is using it.
	  //
	  mb_free (proc_table[pid].address_text,
		   proc_table[pid].domain_text);
	}
    }
  //
  // Abandon children to `init' ((pid_t) 1).
  //
  for (child = 1; child < PROCESS_MAX; child++)
    {
      if (proc_table[child].status != PROC_EMPTY
	  && proc_table[child].ppid == pid)
	{
	  proc_table[child].ppid = 1;	// Son of `init'.
	  if (proc_table[child].status == PROC_ZOMBIE)
	    {
	      sigchld = 1;	// Must send a SIGCHLD to `init'.
	    }
	}
    }
  //
  // SIGCHLD to `init'.
  //
  if (sigchld
      && pid != 1
      && proc_table[1].status != PROC_EMPTY
      && proc_table[1].status != PROC_ZOMBIE)
    {
      proc_sig_on ((pid_t) 1, SIGCHLD);
    }
  //
  // Announce to the parent the death of its child.
  //
  if (pid != parent && proc_table[parent].status != PROC_EMPTY)
    {
      proc_sig_on (parent, SIGCHLD);
    }
}
