#include <errno.h>
#include <kernel/proc.h>
#include <kernel/memory.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
int
s_brk (pid_t pid, void *address)
{
  size_t          requested_size;
  size_t          requested_extra;
  addr_t          previous_address_text;
  size_t          previous_domain_text;
  addr_t          previous_address_data;
  size_t          previous_domain_data;
  size_t          previous_extra;
  addr_t          allocated_text;
  addr_t          allocated_data;
  int             status;
  //
  // All segments start form ((void *) 0), so the new
  // address requested is equivalent to the new size
  // for the data segment.
  //
  requested_size = (size_t) address;
  //
  // Check if it is possible to get the new size: cannot be less
  // then the original segment size.
  //
  if (proc_table[pid].domain_data == 0)
    {
      if (proc_table[pid].domain_text > requested_size)
	{
	  requested_extra = 0;
	}
      else
	{
	  requested_extra = requested_size
	    - proc_table[pid].domain_text;
	}
    }
  else
    {
      if (proc_table[pid].domain_data > requested_size)
	{
	  requested_extra = 0;
	}
      else
	{
	  requested_extra = requested_size
	    - proc_table[pid].domain_data;
	}
    }
  //
  // Now make shure that the new value is a multiple of
  // MEM_BLOCK_SIZE!
  //
  if (requested_extra % MEM_BLOCK_SIZE)
    {
      requested_extra = (((requested_extra / MEM_BLOCK_SIZE) + 1)
			 * MEM_BLOCK_SIZE);
    }
  //
  // Now resize the process.
  //
  if (requested_extra == proc_table[pid].extra_data)
    {
      //
      // Nothing have to change.
      //
      return (0);
    }
  else if (requested_extra < proc_table[pid].extra_data)
    {
      //
      // Just reduce.
      //
      previous_extra = proc_table[pid].extra_data;
      proc_table[pid].extra_data = requested_extra;
      //
      // Update process DATA segment inside the GDT table.
      //
      if (proc_table[pid].domain_data > 0)
	{
	  gdt_segment (gdt_pid_to_segment_data (pid),
		       (uint32_t) proc_table[pid].address_data,
		       (uint32_t) ((proc_table[pid].domain_data
				    + proc_table[pid].extra_data)
				   / MEM_BLOCK_SIZE), 1, 0, 0);
	}
      else
	{
	  gdt_segment (gdt_pid_to_segment_data (pid),
		       (uint32_t) proc_table[pid].address_text,
		       (uint32_t) ((proc_table[pid].domain_text
				    + proc_table[pid].extra_data)
				   / MEM_BLOCK_SIZE), 1, 0, 0);
	}
      //
      // Release memory.
      //
      if (proc_table[pid].domain_data > 0)
	{
	  status = mb_reduce ((proc_table[pid].address_data
			       + proc_table[pid].domain_data),
			      proc_table[pid].extra_data,
			      previous_extra);
	}
      else
	{
	  status = mb_reduce ((proc_table[pid].address_text
			       + proc_table[pid].domain_text),
			      proc_table[pid].extra_data,
			      previous_extra);
	}
      //
      if (status < 0)
	{
	  //
	  // What happened?
	  //
	  k_perror (NULL);
	}
    }
  else
    {
      //
      // A bigger size was requested. Save previous information.
      //
      previous_address_text = proc_table[pid].address_text;
      previous_domain_text = proc_table[pid].domain_text;
      previous_address_data = proc_table[pid].address_data;
      previous_domain_data = proc_table[pid].domain_data;
      previous_extra = proc_table[pid].extra_data;
      //
      // Allocate memory for text,
      // if text and data are inside the same address space;
      // otherwise only the data segment is involved.
      //
      if (proc_table[pid].domain_data == 0)
	{
	  allocated_text = mb_alloc_size (proc_table[pid].domain_text
					  + requested_extra);
	  //
	  if (allocated_text == 0)
	    {
	      errset (ENOMEM);	// Not enough space.
	      return (-1);
	    }
	  //
	  if (DEBUG)
	    {
	      k_printf ("%s:%i:mb_alloc_size(%zi)",
			__FILE__, __LINE__,
			(proc_table[pid].domain_text
			 + requested_extra));
	    }
	}
      //
      // Allocate memory for data, if necessary.
      //
      if (proc_table[pid].domain_data > 0)
	{
	  allocated_data = mb_alloc_size (proc_table[pid].domain_data
					  + requested_extra);
	  //
	  if (allocated_data == 0)
	    {
	      //
	      // Please note that, if we are here, no memory
	      // for the text was allocated!
	      //
	      errset (ENOMEM);	// Not enough space.
	      return (-1);
	    }
	  //
	  if (DEBUG)
	    {
	      k_printf ("%s:%i:mb_alloc_size(%zi)",
			__FILE__, __LINE__,
			(proc_table[pid].domain_data
			 + requested_extra));
	    }
	}
      //
      // Copy the process text and, data in memory: if
      // size is zero, no copy is made. But the text is
      // copied only if text and data live together.
      //
      if (proc_table[pid].domain_data == 0)
	{
	  memcpy ((void *) allocated_text,
		  (void *) proc_table[pid].address_text,
		  (size_t) (proc_table[pid].domain_text
			    + proc_table[pid].extra_data));
	}
      else
	{
	  memcpy ((void *) allocated_data,
		  (void *) proc_table[pid].address_data,
		  (size_t) (proc_table[pid].domain_data
			    + proc_table[pid].extra_data));
	}
      //
      // Update process information.
      //
      if (proc_table[pid].domain_data == 0)
	{
	  proc_table[pid].address_text = allocated_text;
	}
      else
	{
	  proc_table[pid].address_data = allocated_data;
	}
      proc_table[pid].extra_data = requested_extra;
      //
      // Update process TEXT segment inside the GDT table.
      //
      gdt_segment (gdt_pid_to_segment_text (pid),
		   (uint32_t) proc_table[pid].address_text,
		   (uint32_t) (proc_table[pid].domain_text
			       / MEM_BLOCK_SIZE), 1, 1, 0);
      //
      // Update process DATA segment inside the GDT table.
      //
      if (proc_table[pid].domain_data > 0)
	{
	  gdt_segment (gdt_pid_to_segment_data (pid),
		       (uint32_t) proc_table[pid].address_data,
		       (uint32_t) ((proc_table[pid].domain_data
				    + proc_table[pid].extra_data)
				   / MEM_BLOCK_SIZE), 1, 0, 0);
	}
      else
	{
	  gdt_segment (gdt_pid_to_segment_data (pid),
		       (uint32_t) proc_table[pid].address_text,
		       (uint32_t) ((proc_table[pid].domain_text
				    + proc_table[pid].extra_data)
				   / MEM_BLOCK_SIZE), 1, 0, 0);
	}
      //
      // Now release the old memory!
      //
      if (proc_table[pid].domain_data == 0)
	{
	  mb_free (previous_address_text,
		   previous_domain_text + previous_extra);
	}
      else
	{
	  mb_free (previous_address_data,
		   previous_domain_data + previous_extra);
	}
    }
  //
  // Ok.
  //
  return (0);
}
