#include <fcntl.h>
#include <kernel/proc.h>
#include <kernel/net.h>
#include <kernel/net/route.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/os32.h>
#include <netinet/udp.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
ssize_t
s_recvfrom (pid_t pid, int sfdn, void *buffer, size_t length, int flags,
            struct sockaddr *addrfrom, socklen_t *addrlen)
{
  fd_t               *sfd;
  int                 i;          // IP table index.
  size_t              size_read;
  struct udphdr      *udp;
  void               *data;
  struct sockaddr_in *addrfrom_in = (void *) addrfrom;
  //
  // Get file descriptor and verify that it is a socket.
  //
  sfd = fd_reference (pid, &sfdn);
  if (sfd == NULL || sfd->file == NULL)
    {
      errset (EBADF);		// Bad file descriptor.
      return ((ssize_t) -1);
    }
  if (sfd->file->sock == NULL)
    {
      errset (ENOTSOCK);	// Not a socket.
      return ((ssize_t) -1);
    }
  //
  // Verify to have a valid buffer pointer.
  //
  if (buffer == NULL)
    {
      errset (EINVAL);
      return ((ssize_t) -1);
    }
  //
  //
  //
  if (sfd->file->sock->family == AF_INET)                    // INET
    {
      //
      // Should do some check here...
      //
      if (sfd->file->sock->type == SOCK_RAW)                 // RAW
        {
          if (sfd->file->sock->protocol == IPPROTO_ICMP)     // ICMP
            {
              //
              // Scan the ip_table[] to find an ICMP packet
              // that was not already seen by the socket.
              //
              for (i = 0; i < IP_MAX_PACKETS; i++)
                {
                  //
                  // Check the protocol.
                  //
                  if (ip_table[i].packet.header.protocol
                      != IPPROTO_ICMP)
                    {
                      //
                      // It is not ICMP.
                      //
                      continue;
                    }
                  //
                  // Is the packet new for the socket?
                  //
                  // Please notice that the kernel might be interrupted
                  // also between clock tics; so, during a single clock
                  // time, a new packet might be reached.
                  //
                  if (ip_table[i].clock
                      < sfd->file->sock->read.clock[i])
                    {
                      //
                      // Already seen or packet too old.
                      //
                      continue;
                    }
                  //
                  // Verify the IP addresses.
                  //
                  if (ip_table[i].packet.header.daddr
                       != htonl (sfd->file->sock->laddr)
                      && sfd->file->sock->laddr != 0)
                    {
                      //
                      // The local address does not match, and it is
                      // not zero.
                      //
                      continue;
                    }
                  //
                  if (ip_table[i].packet.header.saddr
                       != htonl (sfd->file->sock->raddr))
                    {
                      //
                      // The remote address does not match, but
                      // if it is zero, we accept all.
                      //
                      if (sfd->file->sock->raddr == 0)
                        {
                          //
                          // Can accept the packet.
                          //
                          ;
                        }
                      else
                        {
                          continue;
                        }
                    }
                  //
                  // Packet accepted.
                  //
                  // This ICMP RAW packet is new for the
                  // socket: save the clock time, so that the
                  // same packet is not read again.
                  //
                  sfd->file->sock->read.clock[i]
                   = ip_table[i].clock;
                  //
                  // Copy the packet.
                  //
                  size_read
                   = min (ntohs (ip_table[i].packet.header.tot_len),
                          length);
                  //
                  memcpy (buffer, ip_table[i].packet.octet,
                          size_read);
                  //
                  // Get the source address and return.
                  //
                  if (addrfrom != NULL && addrlen != NULL)
                    {
                      if (*addrlen >= sizeof (struct sockaddr_in))
                        {
                          addrfrom_in->sin_family = AF_INET;
                          addrfrom_in->sin_port = 0;
                          addrfrom_in->sin_addr.s_addr
                           = ip_table[i].packet.header.saddr;
                        }
                      *addrlen = sizeof (struct sockaddr_in);
                    }
                  return ((ssize_t) size_read);
                }
            }
          else
            {
              //
              // Unsupported protocol.
              //
              errset (EPROTONOSUPPORT);
              return ((ssize_t) -1);
            }
        }
      else if (sfd->file->sock->type == SOCK_DGRAM)             // DGRAM
        {
          if (sfd->file->sock->protocol == IPPROTO_UDP)         // UDP
            {
              //
              // Scan the ip_table[] to find an UDP packet
              // that was not already seen by the socket.
              //
              for (i = 0; i < IP_MAX_PACKETS; i++)
                {
                  //
                  // Check the protocol.
                  //
                  if (ip_table[i].packet.header.protocol
                      != IPPROTO_UDP)
                    {
                      //
                      // It is not UDP.
                      //
                      continue;
                    }
                  //
                  // Is the packet new for the socket?
                  //
                  // Please notice that the kernel might be interrupted
                  // also between clock tics; so, during a single clock
                  // time, a new packet might be reached.
                  //
                  if (ip_table[i].clock
                      < sfd->file->sock->read.clock[i])
                    {
                      //
                      // Already seen or packet too old.
                      //
                      continue;
                    }
                  //
                  // Verify the ports.
                  //
                  udp = (struct udphdr *)
                        &ip_table[i].packet.octet
                         [sizeof (struct iphdr)];
                  //
                  if (udp->dest == 0)
                    {
                      //
                      // Cannot accept packets for the port zero!
                      //
                      continue;
                    }
                  //
                  if (udp->dest != htons (sfd->file->sock->lport))
                    {
                      //
                      // The local port does not match!
                      //
                      continue;
                    }
                  //
                  if (udp->source != htons (sfd->file->sock->rport)
                      && sfd->file->sock->rport != 0)
                    {
                      //
                      // The remote port does not match, and is not
                      // zero.
                      //
                      continue;
                    }
                  //
                  // Verify the IP addresses.
                  //
                  if (ip_table[i].packet.header.daddr
                       != htonl (sfd->file->sock->laddr)
                      && sfd->file->sock->laddr != 0)
                    {
                      //
                      // The local address does not match, and is
                      // not zero.
                      //
                      continue;
                    }
                  //
                  if (ip_table[i].packet.header.saddr
                       != htonl (sfd->file->sock->raddr)
                      && sfd->file->sock->raddr != 0)
                    {
                      //
                      // The remote address does not match, and is
                      // not zero.
                      //
                      continue;
                    }
                  //
                  // The packet is accepted.
                  //
                  // This UDP packet is new for the socket:
                  // save the clock time, so that the
                  // same packet is not read again.
                  //
                  sfd->file->sock->read.clock[i]
                   = ip_table[i].clock;
                  //
                  // Check the right minimal size to be read, comparing
                  // the size of the IP packet, the size of the UDP
                  // packet and the size requested.
                  //
                  size_read = ntohs (ip_table[i].packet.header.tot_len)
                               - (ip_table[i].packet.header.ihl * 4)
                               - (sizeof (struct udphdr));
                  size_read = min (size_read,
                                   (udp->len - sizeof (struct udphdr)));
                  size_read = min (size_read, length);
                  //
                  // Copy the data inside the UDP packet.
                  //
                  data = (((uint8_t *) udp) + sizeof (struct udphdr));
                  //
                  memcpy (buffer, data, size_read);
                  //
                  // Get the source address and return.
                  //
                  if (addrfrom != NULL && addrlen != NULL)
                    {
                      if (*addrlen >= sizeof (struct sockaddr_in))
                        {
                          addrfrom_in->sin_family = AF_INET;
                          addrfrom_in->sin_port = udp->source;
                          addrfrom_in->sin_addr.s_addr
                           = ip_table[i].packet.header.saddr;
                        }
                      *addrlen = sizeof (struct sockaddr_in);
                    }
                  return ((ssize_t) size_read);
                }
            }
        }
      else if (sfd->file->sock->type == SOCK_STREAM)            // STREAM
        {
          if (sfd->file->sock->protocol == IPPROTO_TCP)         // TCP
            {
              //
              // See if the read side of the stream was closed.
              //
              if (sfd->file->sock->tcp.recv_closed
                  || sfd->file->sock->tcp.conn == TCP_CLOSE)
                {
                  //
                  // If the `recv_size' is zero, the stream
                  // is closed.
                  //
                  if (sfd->file->sock->tcp.recv_size == 0
                    || sfd->file->sock->tcp.can_read == 0)
                    {
                      //
                      // End of file.
                      //
                      return ((ssize_t) 0);
                    }
                }
              //
              // At the moment, nothing was read.
              //
              size_read = 0;
              //
              // See if there is data to be read from the stream.
              //
              if (sfd->file->sock->tcp.can_read)
                {
                  size_read = min (sfd->file->sock->tcp.recv_size,
                                   length);
                  memcpy (buffer,
                          sfd->file->sock->tcp.recv_index,
                          size_read);
                  //
                  sfd->file->sock->tcp.recv_size -= size_read;
                  sfd->file->sock->tcp.recv_index += size_read;
                  //
                  if (sfd->file->sock->tcp.recv_size == 0)
                    {
                      //
                      // Nothing to be read at the moment.
                      //
                      sfd->file->sock->tcp.can_read = 0;
                      sfd->file->sock->tcp.can_recv = 1;
                    }
                  //
                  // Get the source address and return.
                  //
                  if (addrfrom != NULL && addrlen != NULL)
                    {
                      if (*addrlen >= sizeof (struct sockaddr_in))
                        {
                          addrfrom_in->sin_family = AF_INET;
                          addrfrom_in->sin_port
                           = htons (sfd->file->sock->rport);
                          addrfrom_in->sin_addr.s_addr
                           = htons (sfd->file->sock->raddr);
                        }
                      *addrlen = sizeof (struct sockaddr_in);
                    }
                }
              //
              // Check if something was read.
              //
              if (size_read > 0)
                {
                  //
                  // Return normally.
                  //
                  return ((ssize_t) size_read);
                }
              else
                {
                  //
                  // Nothing to be read at the moment.
                  //
                  if (sfd->fl_flags & O_NONBLOCK)
                    {
                      //
                      // Try again.
                      //
                      errset (EAGAIN);
                      return ((ssize_t) -1);
                    }
                  else
                    {
                      //
                      // Go to sleep and return a temporary error.
                      //
    	              proc_table[pid].status = PROC_SLEEPING;
	              proc_table[pid].ret = 0;
	              proc_table[pid].wakeup_events
	               = WAKEUP_EVENT_SOCK_READ;
	              proc_table[pid].wakeup_sock = sfd->file->sock;
                      if (DEBUG)
                        {
                          k_printf ("[%s:%i] PID %i goes to sleep "
                                    "waiting to receive for a socket.\n",
                                    __FILE__, __LINE__, pid);
                        }
                      //
                      // Nothing was received.
                      //
                      errset (EAGAIN);
                      return ((ssize_t) -1);
                    }
                }
            }
          else
            {
              //
              // Unsupported protocol.
              //
              errset (EPROTONOSUPPORT);
              return ((ssize_t) -1);
            }
        }
      else
        {
          //
          // Unsupported type.
          //
          errset (EPROTONOSUPPORT);
          return ((ssize_t) -1);
        }
    }
  else
    {
      //
      // Unsupported family.
      //
      errset (EAFNOSUPPORT);
      return ((ssize_t) -1);
    }
  //
  // If we are here, there are no more packets to read at the moment.
  //
  if (sfd->fl_flags & O_NONBLOCK)
    {
      //
      // Try again.
      //
      errset (EAGAIN);
      return (-1);
    }
  else
    {
      //
      // The process should go to sleep.
      //
      proc_table[pid].status = PROC_SLEEPING;
      proc_table[pid].ret = 0;
      proc_table[pid].wakeup_events = WAKEUP_EVENT_SOCK_READ;
      proc_table[pid].wakeup_sock = sfd->file->sock;
      if (DEBUG)
        {
          k_printf ("[%s:%i] PID %i goes to sleep "
                    "waiting to receive for a socket.\n",
                    __FILE__, __LINE__, pid);
        }
      //
      // Try again.
      //
      errset (EAGAIN);
      return ((ssize_t) -1);
    }
}
