#include <kernel/proc.h>
#include <kernel/net.h>
#include <kernel/net/route.h>
#include <kernel/net/udp.h>
#include <errno.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/os32.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
ssize_t
s_send (pid_t pid, int sfdn, const void *buffer, size_t size, int flags)
{
  fd_t     *sfd;
  int       status;
  //
  // Get file descriptor and verify that it is a socket.
  //
  sfd = fd_reference (pid, &sfdn);
  if (sfd == NULL || sfd->file == NULL)
    {
      errset (EBADF);		// Bad file descriptor.
      return ((ssize_t) -1);
    }
  if (sfd->file->sock == NULL)
    {
      errset (ENOTSOCK);	// Not a socket.
      return ((ssize_t) -1);
    }
  if (sfd->file->sock->unreach_port)
    {
      errset (ECONNREFUSED);	// Connection refused.
      return ((ssize_t) -1);
    }
  if (sfd->file->sock->unreach_prot)
    {
      errset (ENOPROTOOPT);	// Protocol not available.
      return ((ssize_t) -1);
    }
  if (sfd->file->sock->unreach_host)
    {
      errset (EHOSTUNREACH);	// Host unreachable.
      return ((ssize_t) -1);
    }
  if (sfd->file->sock->unreach_net)
    {
      errset (ENETUNREACH);	// Net unreachable.
      return ((ssize_t) -1);
    }
  //
  // Verify to have a valid buffer pointer.
  //
  if (buffer == NULL)
    {
      errset (EINVAL);
      return ((ssize_t) -1);
    }
  //
  //
  //
  if (sfd->file->sock->family == AF_INET)                    // INET
    {
      //
      // AF_INET requires at least the remote address.
      //
      if (sfd->file->sock->raddr == 0)
        {
          errset (EDESTADDRREQ);
          return ((ssize_t) -1);
        }
      //
      if (sfd->file->sock->type == SOCK_RAW)                 // RAW
        {
          if (sfd->file->sock->protocol == IPPROTO_ICMP)     // ICMP
            {
              status = ip_tx (sfd->file->sock->laddr,
                              sfd->file->sock->raddr,
                              sfd->file->sock->protocol,
                              buffer, size);
              if (status)
                {
                  errset (errno);
                  return ((ssize_t) -1);
                }
              else
                {
                  return ((ssize_t) size);
                }
            }
          else
            {
              //
              // Unsupported protocol.
              //
              errset (EPROTONOSUPPORT);
              return ((ssize_t) -1);
            }
        }
      else if (sfd->file->sock->type == SOCK_DGRAM)          // DGRAM
        {
          if (sfd->file->sock->protocol == IPPROTO_UDP)      // UDP
            {
              status = udp_tx (sfd->file->sock->lport,
                               sfd->file->sock->rport,
                               sfd->file->sock->laddr,
                               sfd->file->sock->raddr,
                               buffer, size);
              if (status)
                {
                  errset (errno);
                  return ((ssize_t) -1);
                }
              else
                {
                  return ((ssize_t) size);
                }
            }
          else
            {
              //
              // Unsupported protocol.
              //
              errset (EPROTONOSUPPORT);
              return ((ssize_t) -1);
            }
        }
      else if (sfd->file->sock->type == SOCK_STREAM)         // STREAM
        {
          if (sfd->file->sock->protocol == IPPROTO_TCP)      // TCP
            {
              //
              // See if the send side of the stream was closed.
              //
              if (sfd->file->sock->tcp.send_closed
                  || sfd->file->sock->tcp.conn == TCP_CLOSE)
                {
                  //
                  // End of file.
                  //
                  if (DEBUG)
                    {
                      k_printf ("end of socket write\n");
                    }
    	          s_kill ((pid_t) 0, pid, SIGPIPE);
		  errset (EPIPE);
		  return ((ssize_t) - 1);
                }
              //
              // Put data to the send buffer, if it is possible.
              //
              if (sfd->file->sock->tcp.can_write)
                {
                  size = min (size, (TCP_MSS - sizeof (struct tcphdr)));
                  memcpy (sfd->file->sock->tcp.send_data,
                          buffer, size);
                  sfd->file->sock->tcp.send_size = size;
                  sfd->file->sock->tcp.can_write = 0;
                  sfd->file->sock->tcp.can_send  = 1;
                  //
                  sfd->file->sock->tcp.lsq[++sfd->file->sock->tcp.lsqi]
                   = sfd->file->sock->tcp.lsq_ack;
                  sfd->file->sock->tcp.send_flags
                   = TCP_FLAG_PSH|TCP_FLAG_ACK;
                  tcp_tx_sock (sfd->file->sock);
                  //
                  return ((ssize_t) size);
                }
              else
                {
                  //
                  // At the moment, nothing can be written.
                  //
                  if (sfd->fl_flags & O_NONBLOCK)
                    {
                      //
                      // Cannot block.
                      //
                      errset (EAGAIN);
                      return ((ssize_t) -1);
                    }
                  else
                    {
                      //
                      // Go to sleep and return zero.
                      //
    	              proc_table[pid].status = PROC_SLEEPING;
    	              proc_table[pid].ret = 0;
	              proc_table[pid].wakeup_events
	               = WAKEUP_EVENT_SOCK_WRITE;
	              proc_table[pid].wakeup_sock = sfd->file->sock;
                      if (DEBUG)
                        {
                          k_printf ("[%s] PID %i goes to sleep "
                                    "waiting to write to a socket.\n",
                                    __FILE__, pid);
                        }
                      //
                      // Retry.
                      //
                      errset (EAGAIN);
                      return ((ssize_t) -1);
                    }
                }
            }
          else
            {
              //
              // Unsupported protocol.
              //
              errset (EPROTONOSUPPORT);
              return ((ssize_t) -1);
            }
        }
      else
        {
          //
          // Unsupported type.
          //
          errset (EPROTONOSUPPORT);
          return ((ssize_t) -1);
        }
    }
  else
    {
      //
      // Unsupported family.
      //
      errset (EAFNOSUPPORT);
      return ((ssize_t) -1);
    }
}
