#include <stdlib.h>
#include <string.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <kernel/net.h>
#include <kernel/net/tcp.h>
#include <kernel/fs.h>
#include <kernel/lib_s.h>
#include <kernel/lib_k.h>
#include <errno.h>
//----------------------------------------------------------------------
#define DEBUG 0
//----------------------------------------------------------------------
int
tcp (void)
{
  int                s;                        // Socket table index.
  int                p;                        // IP table index.
  int                q;                        // Queue index.
  int                status;
  struct tcphdr     *tcp;
  struct iphdr      *ip;
  int                sfdn;                     // New socket.
  fd_t              *sfd;
  sock_t            *sock;
  struct sockaddr_in sa;
  clock_t            delay;
  uint8_t           *recv_data;
  size_t             recv_size;
  int                ret = 0;
  unsigned int       lseq;
  //
  // Scan local sockets.
  //
  for (s = 0; s < SOCK_MAX_SLOTS; s++)
    {
      if (!sock_table[s].active)                 continue;
      if (sock_table[s].family != AF_INET)       continue;
      if (sock_table[s].protocol != IPPROTO_TCP) continue;
      if (sock_table[s].unreach_port)            continue;
      if (sock_table[s].unreach_host)            continue;
      //
      // Calculate the delay from the last send.
      //
      delay = s_clock ((pid_t) 0) - sock_table[s].tcp.clock;
      //
      // Have we received something? Scan the ip_table[] to find a
      // TCP packet that was not already seen by the socket.
      //
      for (p = 0; p < IP_MAX_PACKETS; p++)
        {
          //////////////////////////////////////////////////////////////
          // PACKET CHECK
          //////////////////////////////////////////////////////////////
          //
          // Check the protocol.
          //
          if (ip_table[p].packet.header.protocol != IPPROTO_TCP)
            {
              //
              // It is not TCP.
              //
              continue;
            }
          //
          // Is the packet new for the socket?
          //
          if (ip_table[p].clock <= sock_table[s].read.clock[p])
            {
              //
              // Already seen or packet too old.
              //
              continue;
            }
          //
          // Get a pointer to IP and TCP headers.
          //
          ip  = (struct iphdr *) &ip_table[p].packet.header;
          tcp = (struct tcphdr *)
                &ip_table[p].packet.octet[ip->ihl * 4];
          //
          // Verify the ports.
          //
          if (tcp->dest != htons (sock_table[s].lport))
            {
              //
              // The local port does not match!
              //
              continue;
            }
            //
          if (tcp->source != htons (sock_table[s].rport))
            {
              //
              // The remote port does not match, but might be
              // listening and the packet might be a SYN.
              //
              if (sock_table[s].rport == 0
                  && sock_table[s].tcp.conn == TCP_LISTEN
                  && tcp->syn
                  && !tcp->ack)
                {
                  //
                  // We hope that it is the first SYN.
                  //
                  ;
                }
              else
                {
                  continue;
                }
            }
          //
          // Verify the IP addresses.
          //
          if (ip_table[p].packet.header.daddr
               != htonl (sock_table[s].laddr))
            {
              //
              // The local address does not match, but
              // might be zero.
              //
              if (sock_table[s].laddr != 0)
                {
                  //
                  // The local address is not zero, so
                  // the match fails.
                  //
                  continue;
                }
            }
          //
          if (ip_table[p].packet.header.saddr
               != htonl (sock_table[s].raddr))
            {
              //
              // The remote address does not match, but the socket
              // might be listening and che packet might be a SYN.
              //
              if (sock_table[s].raddr == 0
                  && sock_table[s].tcp.conn == TCP_LISTEN
                  && tcp->syn
                  && !tcp->ack)
                {
                  //
                  // We hope that it is the first SYN.
                  //
                  ;
                }
              else
                {
                  continue;
                }
            }
          //
          // This TCP packet is new for the socket:
          // save the clock time, so that the
          // same packet is not read again.
          //
          sock_table[s].read.clock[p] = ip_table[p].clock;
          //
          //////////////////////////////////////////////////////////////
          // TCP PROTOCOL
          //////////////////////////////////////////////////////////////
          //
          if (DEBUG)
            {
              tcp_show (ntohl (ip->saddr), ntohl (ip->daddr), tcp);
            }
          //
          recv_data = &((uint8_t *) tcp)[tcp->doff * 4];
          recv_size = ntohs (ip->tot_len) - (ip->ihl * 4)
                                          - (tcp->doff * 4);
          //
          // Now we have received a TCP packet for the current
          // socket, and we should do something with it...
          //
          if (tcp->rst)
            {
              //
              // We have received a reset... What is resetting?
              //
              if ((sock_table[s].tcp.rsq[sock_table[s].tcp.rsqi]
                   == ntohl (tcp->seq))
                  || (tcp->ack
                      && (sock_table[s].tcp.lsq_ack
                          == ntohl (tcp->ack_seq))))
                {
                  sock_table[s].tcp.recv_closed = 1;
                  sock_table[s].tcp.send_closed = 1;
                  sock_table[s].tcp.conn = TCP_RESET;
                  if (DEBUG)
                    {
                      k_printf ("[%s] TCP_RESET\n", __func__);
                    }
                }
              else
                {
                  k_printf ("lsq_ack=%3, rsq=%3\n",
                            sock_table[s].tcp.lsq_ack,
                            sock_table[s].tcp.rsq[sock_table[s].tcp.rsqi]);
                }
            }
          else if (sock_table[s].tcp.conn == 0
                   || sock_table[s].tcp.conn == TCP_CLOSE
                   || sock_table[s].tcp.conn == TCP_RESET)
            {
              //
              // The connection is not yet ready or it is closed.
              // We are not waiting any packet, so we just reject it.
              //
              tcp_tx_rst (ip);
            }
          else if (sock_table[s].tcp.conn == TCP_LISTEN)
            {
              //
              // It should be a first SYN packet.
              //
              if (tcp->syn && !tcp->ack)
                {
                  //
                  // Should be a new connection attempt. Can we queue
                  // it?
                  //
                  for (q = 0; q < sock_table[s].tcp.listen_max; q++)
                    {
                      if (sock_table[s].tcp.listen_queue[q] == -1)
                        {
                          break;
                        }
                    }
                  if (q >= sock_table[s].tcp.listen_max)
                    {
                      //
                      // The queue is full.
                      //
                      tcp_tx_rst (ip);
                      //
                      // Next packet.
                      //
                      continue;
                    }
                  //
                  // Is this connection attempt already done?
                  //
                  status = tcp_status (ip);
                  //
                  if (status < 0)
                    {
                      //
                      // Should not happen.
                      //
                      errset (errno);
                      perror (NULL);
                      //
                      // Ignore the packet?!
                      //
                      continue;
                    }
                  else if (status == TCP_SYN_SENT)
                    {
                      //
                      // The same SYN was already received and serviced:
                      // just ignore the packet.
                      //
                      continue;
                    }
                  else if (status > 0)
                    {
                      //
                      // There is already a connection with the same
                      // addresses: ignore the SYN.
                      //
                      continue;
                    }
                  //
                  // The SYN is new!
                  // Can we open a new socket?
                  //
                  sfdn = s_socket (sock_table[s].tcp.listen_pid,
                                   AF_INET, SOCK_STREAM, IPPROTO_TCP);
                  if (sfdn < 0)
                    {
                      //
                      // No, sorry.
                      //
                      tcp_tx_rst (ip);
                      //
                      // Next packet.
                      //
                      continue;
                    }
                  //
                  // Can we bind it to the same destination of the
                  // received packet?
                  //
                  sa.sin_family      = AF_INET;
                  sa.sin_port        = tcp->dest;
                  sa.sin_addr.s_addr = ip->daddr;
                  status = s_bind (sock_table[s].tcp.listen_pid,
                                   sfdn, (struct sockaddr *) &sa,
                                   sizeof (sa));
                  if (status < 0)
                    {
                      //
                      // No, sorry.
                      //
                      tcp_tx_rst (ip);
                      close (sfdn);
                      //
                      // Next packet.
                      //
                      continue;
                    }
                  //
                  // Ok. Save the new socket number in queue.
                  //
                  sock_table[s].tcp.listen_queue[q] = sfdn;
                  //
                  // Prepare some pointers to reach the new socket
                  // easily.
                  //
                  sfd = fd_reference (sock_table[s].tcp.listen_pid,
                                      &sfdn);
                  sock = sfd->file->sock;
                  //
                  // Connect the new socket with the remote node.
                  //
                  sock->raddr = ntohl (ip->saddr);
                  sock->rport = ntohs (tcp->source);
                  //
                  // The new socket has seen this SYN packet.
                  //
                  sock->read.clock[p] = ip_table[p].clock;
                  //
                  // Make the new socket answare with a second
                  // SYN.
                  //
                  memset (sock->tcp.lsq, 0x00, sizeof (sock->tcp.lsq));
                  memset (sock->tcp.rsq, 0x00, sizeof (sock->tcp.rsq));
                  srand ((unsigned int) s_clock ((pid_t) 0));
                  lseq = rand ();
                  sock->tcp.lsq_ack = lseq +1;
                  sock->tcp.lsq[++sock->tcp.lsqi] = lseq;
                  sock->tcp.rsq[++sock->tcp.rsqi] = ntohl (tcp->seq);
                  sock->tcp.rsq[++sock->tcp.rsqi] = ntohl (tcp->seq) +1;
                  //
                  sock->tcp.can_send    = 1;
                  sock->tcp.send_flags  = TCP_FLAG_SYN|TCP_FLAG_ACK;
                  if (DEBUG)
                    {
                      k_printf ("[%s] New conn. seq=%3u, lsq_ack=%3u\n",
                                __func__, lseq, sock->tcp.lsq_ack);
                    }
                  tcp_tx_sock (sock);
                  //
                  // Put the new socket to TCP_SYN_RECV status.
                  //
                  sock->tcp.conn = TCP_SYN_RECV;
                }
              else
                {
                  //
                  // We are listening: cannot accept other type of packets.
                  //
                  tcp_tx_rst (ip);
                }
            }
          else if (sock_table[s].tcp.conn == TCP_SYN_SENT)
            {
              //
              // It should be a second SYN packet with ACK.
              //
              if (tcp->syn
                  && tcp->ack
                  && tcp_rx_ack (&sock_table[s], ip) == 0)
                {
                  //
                  // SYN + ACK.
                  //
                  // Save the initial remote sequence, because it
                  // is the first one, and save also the remote
                  // sequence that we will expect next time.
                  //
                  sock_table[s].tcp.rsq[++sock_table[s].tcp.rsqi]
                   = ntohl (tcp->seq);
                  sock_table[s].tcp.rsq[++sock_table[s].tcp.rsqi]
                   = ntohl (tcp->seq) + 1;
                  //
                  // The received SYN is to be confirmed with ACK.
                  // The expected next local sequence does not change,
                  // and in effect, we don't expect any other ACK back.
                  //
                  sock_table[s].tcp.lsq[++sock_table[s].tcp.lsqi]
                   = sock_table[s].tcp.lsq_ack;
                  tcp_tx_ack (&sock_table[s]);
                  //
                  // We are now in TCP_ESTABLISHED.
                  //
                  sock_table[s].tcp.conn = TCP_ESTABLISHED;
                  //
                  // Now the process can write and can receive
                  // data (can write --but cannot send-- and can
                  // receive --but cannot read--).
                  //
                  sock_table[s].tcp.can_write = 1;
                  sock_table[s].tcp.can_send  = 0;
                  //
                  sock_table[s].tcp.can_recv = 1;
                  sock_table[s].tcp.can_read = 0;
                  //
                  ret |= TCP_TRY_WRITE;
                }
              //
              // The case of a single ACK and a single SYN is not
              // taken into consideration!
              //
              // No other type of packet is expected here.
              //
            }
          else if (sock_table[s].tcp.conn == TCP_SYN_RECV)
            {
              //
              // We are waiting an ACK for our second SYN.
              //
              if (tcp->ack && tcp_rx_ack (&sock_table[s], ip) == 0)
                {
                  //
                  // ACK ok.
                  // The connection is ready.
                  //
                  sock_table[s].tcp.conn = TCP_ESTABLISHED;
                  //
                  // Now the process can write and can receive
                  // data (can write --but cannot send-- and can
                  // receive --but cannot read--).
                  //
                  sock_table[s].tcp.can_write = 1;
                  sock_table[s].tcp.can_send  = 0;
                  //
                  sock_table[s].tcp.can_recv = 1;
                  sock_table[s].tcp.can_read = 0;
                  //
                  ret |= TCP_TRY_WRITE;
                }
              //
              // No other type of packet is expected here.
              //
            }
          else if (sock_table[s].tcp.conn == TCP_ESTABLISHED)
            {
              //
              // Might be a repeated SYN + ACK, because the other
              // side don't have received our ACK. Just resend.
              //
              if (tcp->syn && tcp->ack)
                {
                  tcp_tx_ack (&sock_table[s]);
                  //
                  // Next packet.
                  //
                  continue;
                }
              //
              // It might be a normal ACK.
              //
              if (tcp->ack)
                {
                  //
                  // Verify if the packet contains data: if there is data,
                  // before sending the ACK, must verify to be able to receive
                  // such data.
                  //
                  if (recv_size > 0)
                    {
                      // k_printf ("[%i]", (int) recv_size);
                      //
                      // There is data.
                      //
                      if (!sock_table[s].tcp.can_recv)
                        {
                          //
                          // At the moment, cannot receive: the packet
                          // is currently ignored and no ACK is sent.
                          //
                          continue;
                        }
                    }
                  //
                  // The received packet is empty or it can be received.
                  //
                  if (tcp_rx_ack (&sock_table[s], ip) == 0)
                    {
                      //
                      // ACK ok.
                      // The process can continue to write and the packet
                      // don't have to be resent.
                      //
                      sock_table[s].tcp.can_write = 1;
                      sock_table[s].tcp.can_send  = 0;
                      //
                      ret |= TCP_TRY_WRITE;
                    }
                  else
                    {
                      //
                      // Next packet.
                      //
                      continue;
                    }
                }
              //
              // It might be a FIN.
              //
              if (tcp->fin)
                {
                  //
                  // Is the FIN in the right sequence?
                  //
                  if (sock_table[s].tcp.rsq[sock_table[s].tcp.rsqi]
                      == ntohl (tcp->seq))
                    {
                      //
                      // Yes, it is: close receiving.
                      //
                      sock_table[s].tcp.recv_closed = 1;
                      //
                      // ACK.
                      //
                      sock_table[s].tcp.rsq[++sock_table[s].tcp.rsqi]
                       = ntohl (tcp->seq) + 1;
                      sock_table[s].tcp.lsq[++sock_table[s].tcp.lsqi]
                       = sock_table[s].tcp.lsq_ack;
                      tcp_tx_ack (&sock_table[s]);
                      //
                      // Change status.
                      //
                      if (DEBUG)
                        {
                          k_printf ("[%s] TCP_CLOSE_WAIT\n", __func__);
                        }
                      sock_table[s].tcp.conn = TCP_CLOSE_WAIT;
                    }
                  else
                    {
                      //
                      // Just ignore it and jump to the next packet.
                      //
                      continue;
                    }
                }
              //
              // The received packet might contain some data, but can
              // accept data only if the receiving buffer is empty
              // (was already read from the reading process).
              //
              tcp_rx_data (&sock_table[s], ip);
              //
              ret |= TCP_TRY_READ;
            }
          else if (sock_table[s].tcp.conn == TCP_CLOSE_WAIT)
            {
              //
              // It might be an ACK.
              //
              if (tcp->ack)
                {
                  if (tcp_rx_ack (&sock_table[s], ip) == 0)
                    {
                      //
                      // ACK ok.
                      // The process can continue to write and the packet
                      // don't have to be resent.
                      //
                      sock_table[s].tcp.can_write = 1;
                      sock_table[s].tcp.can_send  = 0;
                      //
                      ret |= TCP_TRY_WRITE;
                    }
                  else
                    {
                      //
                      // Next packet.
                      //
                      continue;
                    }
                }
              //
              // The data coming from the outside is not taken anymore.
              //
            }
          else if (sock_table[s].tcp.conn == TCP_LAST_ACK)
            {
              //
              // It might be the final ACK.
              //
              if (tcp->ack)
                {
                  if (tcp_rx_ack (&sock_table[s], ip) == 0)
                    {
                      //
                      // ACK ok. The two directions are closed and
                      // the packet confirmed don't have to be resent.
                      //
                      sock_table[s].tcp.recv_closed = 1;
                      sock_table[s].tcp.send_closed = 1;
                      sock_table[s].tcp.can_send    = 0;
                      //
                      // Change status.
                      //
                      if (DEBUG)
                        {
                          k_printf ("[%s] TCP_CLOSE\n", __func__);
                        }
                      sock_table[s].tcp.conn = TCP_CLOSE;
                    }
                  else
                    {
                      //
                      // Next packet.
                      //
                      continue;
                    }
                }
              //
              // The data coming from the outside is not taken anymore.
              //
            }
          else if (sock_table[s].tcp.conn == TCP_FIN_WAIT1)
            {
              if (tcp->ack && tcp->fin)
                {
                  //
                  // ACK and FIN
                  //
                  if (tcp_rx_ack (&sock_table[s], ip) == 0)
                    {
                      //
                      // ACK ok: the confirmed packet don't have to
                      // be resent.
                      //
                      sock_table[s].tcp.conn = TCP_FIN_WAIT2;
                      sock_table[s].tcp.can_send = 0;
                    }
                  else
                    {
                      //
                      // Next packet.
                      //
                      continue;
                    }
                  //
                  // Is the FIN in the right sequence?
                  //
                  if (sock_table[s].tcp.rsq[sock_table[s].tcp.rsqi]
                      == ntohl (tcp->seq))
                    {
                      //
                      // Yes, it is: close receiving.
                      //
                      sock_table[s].tcp.recv_closed = 1;
                      //
                      // ACK.
                      //
                      sock_table[s].tcp.rsq[++sock_table[s].tcp.rsqi]
                       = ntohl (tcp->seq) + 1;
                      sock_table[s].tcp.lsq[++sock_table[s].tcp.lsqi]
                       = sock_table[s].tcp.lsq_ack;
                      tcp_tx_ack (&sock_table[s]);
                      //
                      // Change status.
                      //
                      if (DEBUG)
                        {
                          k_printf ("[%s] TCP_TIME_WAIT\n", __func__);
                        }
                      sock_table[s].tcp.conn = TCP_TIME_WAIT;
                    }
                  else
                    {
                      //
                      // Just ignore it and jump to the next packet.
                      //
                      continue;
                    }
                }
              else if (tcp->ack)
                {
                  //
                  // ACK only.
                  //
                  if (tcp_rx_ack (&sock_table[s], ip) == 0)
                    {
                      //
                      // ACK ok: the confirmed packet don't have to
                      // be resent.
                      //
                      sock_table[s].tcp.conn = TCP_FIN_WAIT2;
                      sock_table[s].tcp.can_send = 0;
                    }
                  else
                    {
                      //
                      // Next packet.
                      //
                      continue;
                    }
                }
              //
              // The received packet might contain some data, but can
              // accept data only if the receive channel is open and
              // if the receiving buffer is empty.
              //
              tcp_rx_data (&sock_table[s], ip);
              //
              ret |= TCP_TRY_READ;
            }
          else if (sock_table[s].tcp.conn == TCP_FIN_WAIT2)
            {
              //
              // It might be the final ACK.
              //
              if (tcp->fin && tcp->ack)
                {
                  if (tcp_rx_ack (&sock_table[s], ip) == 0)
                    {
                      //
                      // ACK ok: the packet don't have to be resent.
                      //
                      sock_table[s].tcp.conn = TCP_TIME_WAIT;
                      sock_table[s].tcp.can_send = 0;
                      //
                      // Next packet.
                      //
                      continue;
                    }
                  else
                    {
                      //
                      // Next packet.
                      //
                      continue;
                    }
                }
              //
              // The received packet might contain some data, but can
              // accept data only if the receiving buffer is empty
              // (was already read from the reading process).
              //
              tcp_rx_data (&sock_table[s], ip);
              //
              ret |= TCP_TRY_READ;
            }
          else if (sock_table[s].tcp.conn == TCP_TIME_WAIT)
            {
              //
              // It might be duplicate final ACK.
              //
              if (tcp->fin && tcp->ack)
                {
                  //
                  // Just resend the ACK.
                  //
                  tcp_tx_ack (&sock_table[s]);
                }
              //
              // Close receiving too, if it is not already done.
              //
              sock_table[s].tcp.recv_closed = 1;
              //
              // Change status, without waiting anymore.
              //
              if (DEBUG)
                {
                  k_printf ("[%s] TCP_CLOSE\n", __func__);
                }
              sock_table[s].tcp.conn = TCP_CLOSE;
            }
        }
      //
      // See if there is something to be re-sent (if the flag
      // `tcp.can_send' is set).
      //
      if (sock_table[s].tcp.can_send && delay > TCP_MAX_DELAY)
        {
          tcp_tx_sock (&sock_table[s]);
        }
    }
  //
  // Return.
  //
  return (ret);
}
