#include <kernel/proc.h>
#include <kernel/lib_k.h>
#include <kernel/lib_s.h>
#include <kernel/net.h>
#include <stdint.h>
//----------------------------------------------------------------------
#define DEBUG 1
//----------------------------------------------------------------------
extern uint32_t _ksp;
extern uint32_t proc_stack_pointer;
extern uint16_t proc_stack_segment_selector;
extern pid_t    proc_current;
//----------------------------------------------------------------------
void
proc_scheduler (void)
{
    pid_t    prev;
    pid_t    next;
    addr_t   stack_top;
    addr_t   stack_bottom;
    uint32_t saved_stack_pointer = proc_stack_pointer;
    //
    static unsigned long long int previous_clock;
           unsigned long long int current_clock;
    //
    // Check the current stack size.
    //
    if (proc_table[proc_current].domain_data == 0)
      {
        stack_bottom = proc_table[proc_current].domain_text;
      }
    else
      {
        stack_bottom = proc_table[proc_current].domain_data;
      }
    //
    stack_top = stack_bottom - proc_table[proc_current].domain_stack;
    //
    // Check if the process has broken data with the stack,
    // or if it is near the end of its domain.
    //
    if (proc_stack_pointer <= stack_top)
      {
        //
        // The stack overlaped the other data!
        //
        k_printf ("[%s] Kernel alert: the stack of process %i "
                  "is grown beyond the allowed space! The process "
                  "is closed. Stack top is %i, stack pointer is %i.\n",
                  __FILE__, (int) proc_current, (int) stack_top,
                  (int) proc_stack_pointer);
        //
        // The process is terminated badly.
        //
        s__exit (proc_current, -1);
      }
    else if (proc_stack_pointer < (stack_top + 1024))
      {
        //
        // There is only 1 Kibyte and the stack is finished!
        //
        k_printf ("[%s] Kernel alert: the stack of process %i "
                  "is near the end of the allowed space! "
                  "It remains only %i byte and it will "
                  "overwrite other data!\n",
                  __FILE__, (int) proc_current,
                  (int) (proc_stack_pointer - stack_top));
      }
    //
    // Save previous PID. Variable `proc_current' is extern.
    //
    prev = proc_current;
    //
    // Take care of networking.
    //
    proc_sch_net ();
    //
    // Take care of sleeping processes: wake up if sleeping time
    // elapsed.
    //
    proc_sch_timers ();
    //
    // Take care of pending signals.
    //
    proc_sch_signals ();
    //
    // Take care input from terminals.
    //
    proc_sch_terminals ();
    //
    // Update the CPU time usage.
    //
    current_clock   = s_clock ((pid_t) 0);
    proc_table[prev].usage += current_clock - previous_clock;
    previous_clock  = current_clock;
    //
    // Check stack pointer changes, made probably by
    // `proc_sig_handler()' called from `proc_sch_signals()'.
    //
    if (DEBUG)
      {
        if (saved_stack_pointer != proc_stack_pointer)
          {
            k_printf ("[%s] pid %i, ESP from %i to %i.\n",
                      __FILE__, proc_current,
                      saved_stack_pointer, proc_stack_pointer);
          }
      }
    //
    // Scan for a next process.
    //
    for (next = prev+1; next != prev; next++)
      {
        if (next >= PROCESS_MAX)
          {
            next = -1; // At the next loop, `next' will be zero.
            continue;
          }
        //
        if      (proc_table[next].status == PROC_EMPTY)
          {
            continue;
          }
        else if (proc_table[next].status == PROC_CREATED) 
          {
            continue;
          }
        else if (proc_table[next].status == PROC_READY) 
          {
            if (proc_table[prev].status == PROC_RUNNING)
              {
                proc_table[prev].status = PROC_READY;
              }
            //
            proc_table[prev].sp     = proc_stack_pointer;
            proc_table[next].status = PROC_RUNNING;
            proc_table[next].ret    = 0;
            //
            proc_current            = next;
            proc_stack_segment_selector
                = gdt_pid_to_segment_data (next) * 8;
            proc_stack_pointer      = proc_table[next].sp;
            break;
          }
        else if (proc_table[next].status == PROC_RUNNING)
          {
            if (proc_table[prev].status == PROC_RUNNING)
              {
                k_printf ("Kernel alert: process %i "
                          "and %i \"running\"!\n",
                          prev, next);
                proc_table[prev].status = PROC_READY;
              }
            //
            proc_table[prev].sp     = proc_stack_pointer;
            proc_table[next].status = PROC_RUNNING;
            proc_table[next].ret    = 0;
            //
            proc_current            = next;
            proc_stack_segment_selector
                = gdt_pid_to_segment_data (next) * 8;
            proc_stack_pointer      = proc_table[next].sp;
            break;
          }
        else if (proc_table[next].status == PROC_SLEEPING)
          {
            continue;
          }
        else if (proc_table[next].status == PROC_ZOMBIE)
          {
            continue;
          }
      }
    //
    // Check again if the next process is set to running, otherwise set
    // the kernel to such value!
    //
    if (proc_table[next].status != PROC_RUNNING)
      {
        proc_table[0].status  = PROC_RUNNING;
        proc_current          = 0;
        proc_stack_segment_selector
            = gdt_pid_to_segment_data (0) * 8;
        proc_stack_pointer    = proc_table[0].sp;
      }
    //
    // Save kernel stack pointer.
    //
    _ksp = proc_table[0].sp;
}
